function [z,E,z_,E_,gn_,rM_,sM_,r_,tol_] = MDS(z,DEL,eE,eEd,ez,TM,verb)

  I = ones(size(DEL), class(DEL));
  missing = find(isnan(DEL));
  if ~isempty(missing),  I(missing) = 0;  end
  zeroWgts = find(I == 0);
  if ~isempty(zeroWgts), DEL(zeroWgts) = 1;    end
  
  
  E_1 = inf;
  
  sM=10;
  r=.5;
  p=0.5;
  t = 1; 
  searchitermax=0;
  if nargout>2,   
      z_(:,t)=z;      
      rM_(t)=NaN;     
      sM_(t)=NaN;     
      r_(t)=NaN;      
      tol_(t)=NaN;    
  end;  

  %fstressgrad=@fstressgrad_ads;
  %fstressgrad=@fstressgrad_rds;
  fstressgrad=@fstressgrad_sam;
  

  
  while true
      
      [E gE]=fstressgrad(z, DEL, I); 

      gn=norm(gE,inf);
      rM=tanh(sM/2)/gn;
      
     
      
      if nargout > 2, 
         E_(t)=E;       
         gn_(t)=gn;     
      end;
      
      if     E     < eE,  if verb>0, disp('exit: STRESS <eE              ');end; break;
      elseif E_1-E < eEd, if verb>0, disp('exit: DELTA E <eEd            '); if E_1<E, disp('STRES GROWS'); end; end; break;
      elseif gn    < eE,  if verb>0, disp('exit: INF NORM OF GRADIENT <eE');end; break;
      elseif rM    < ez,  if verb>0, disp('exit: rM <ez                  ');end; break;
      elseif t     > TM,  if verb>0, disp('exit: ITERATION LIMIT         ');end; break;
      end
      
      E_1 = E;
      
      q = @(r) fstressgrad( M(-r*gE, z), DEL, I );
      
      %dq0=(q(1e-6)-q(0))/1e-6;
      
      M1t=M1(z,gE,0);        dq0=dot([real(gE); imag(gE)], [real(M1t); imag(M1t)]);
   
      
      [r searchiter]= bhlsearch(q,dq0,p,rM,r);
      searchitermax=max(searchitermax,searchiter);
      
      z = M(-r*gE,z);
      
      t = t+1;
      if nargout > 2,    
          z_(:,t)=z;    
          rM_(t)=rM;    
          sM_(t)=sM;              
          r_(t)=r;      
          tol_(t)=NaN;
      end;
     
  end
      
  if verb > 0,  
     disp(sprintf('%d points %d iterations, E = %.12g',length(z) ,t,E)); 
  end




%______________________________________________________________
% ABSOLUTE DIFFERENCES SQUARED
function [E gE]=fstressgrad_ads(z,DEL,I)
  n = length(z);
  D = hdistmat(z);
  ridge = diag(ones(n,1));
  E =  sum(sum( triu( I .* (D - DEL) ,1).^2 ) );  
  
  if nargout>1,
         gE=zeros(n,1);   %pre-allocation saves time, also this ensures gE is a column
         
         for r=1:n,
               dE1=0; for k=1:(r-1),
                  dE1=dE1+  I(r,k)* ( D(r,k)-DEL(r,k) ) * (dddy(z(r),z(k),r,k,r,1)+i*dddy(z(r),z(k),r,k,r,2));
               end; 
               
               dE2=0; for k=(r+1):n,
                  dE2=dE2+  I(r,k)* ( D(r,k)-DEL(r,k) ) * (dddy(z(k),z(r),k,r,r,1)+i*dddy(z(k),z(r),k,r,r,2));
               end
               
               gE(r)=2*(dE1+dE2);
         end;
  end 


%______________________________________________________________
% RELATIVE DIFFERENCES SQUARED
function [E gE]=fstressgrad_rds(z,DEL,I)
  n = length(z);
  D = hdistmat(z);
  ridge = diag(ones(n,1));
  E = sum(sum( triu( I .* (D - DEL) ./ (DEL+ridge), 1 ).^2 ) );
  
  if nargout>1,
         gE=zeros(n,1);   %pre-allocation saves time, also this ensures gE is a column
         
         for r=1:n,
               dE1=0; for k=1:(r-1),
                  dE1=dE1+  I(r,k)* ( D(r,k)-DEL(r,k) )/(DEL(r,k))^2 * (dddy(z(r),z(k),r,k,r,1)+i*dddy(z(r),z(k),r,k,r,2));
               end; 
               
               dE2=0; for k=(r+1):n,
                  dE2=dE2+  I(r,k)* ( D(r,k)-DEL(r,k) )/(DEL(r,k))^2 * (dddy(z(k),z(r),k,r,r,1)+i*dddy(z(k),z(r),k,r,r,2));
               end
               
               gE(r)=2*(dE1+dE2);
         end;
  end 


%______________________________________________________________
% SAMMON CRITERION

function [E gE]=fstressgrad_sam(z,DEL,I)
  
  
  n = length(z);
  D = hdistmat(z);
  ridge = diag(ones(n,1));
  c=sum(sum(triu(I.*DEL,1)));
  
  E =  sum(sum( triu( I .* (D - DEL).^2 ./ (DEL+ridge), 1) ) )   / c;  
  
  
  if nargout>1,
         gE=zeros(n,1);   %pre-allocation saves time, also this ensures gE is a column
         
         for r=1:n,
               dE1=0; for k=1:(r-1),
                  dE1=dE1+  I(r,k)*(D(r,k)-DEL(r,k))/DEL(r,k) * (dddy(z(r),z(k),r,k,r,1)+i*dddy(z(r),z(k),r,k,r,2));
               end; 
               
               dE2=0; for k=(r+1):n,
                  dE2=dE2+  I(r,k)*(D(r,k)-DEL(r,k))/DEL(r,k) * (dddy(z(k),z(r),k,r,r,1)+i*dddy(z(k),z(r),k,r,r,2));
               end
               
               gE(r)=2*(dE1+dE2)/c;
         end;
  end 




%______________________________________________________________
function dddy=dddy(zj,zk,j,k,r,s)
  %disp([zj,zk,j,k,r,s]);
  
  yj1=real(zj);  yj2=imag(zj);
  yk1=real(zk);  yk2=imag(zk);
  m=sqrt( (yj1-yk1)^2 +  (yj2-yk2)^2 );
  l=sqrt( (yj1^2+yj2^2)*(yk1^2+yk2^2) -2*yj1*yk1 -2*yj2*yk2 +1 );
  
  
  if      r==j,
    if s==1, 
     md= ( yj1 - yk1 )/m;
     ld=g(yj1,yk1,yk2)/l;     
    else 
     md= ( yj2 - yk2 )/m;
     ld=g(yj2,yk2,yk1)/l;
    end;
  elseif  r==k,
    if s==1, 
     md=-( yj1 - yk1 )/m;     
     ld=g(yk1,yj1,yj2)/l;
    else
     md=-( yj2 - yk2 )/m;
     ld=g(yk2,yj2,yj1)/l;
    end;
  else                  
     md=0;
     ld=0;
  end;
  
  dddy=2*(m*ld - l*md) / ( m^2 - l^2 );
  %error('stop');

%______________________________________________________________
function g=g(x1,x2,x3)  
  g=x1*(x2^2 + x3^2)-x2;
  

